import os
from time import time

import casadi
import numpy as np
import pinocchio as pin
from .ProblemData import ProblemData
from .ShootingNode import ShootingNode
from .Target import Target
import pinocchio.casadi as cpin


class NodeData:
    """Data stored by each shooting node of the OCP

    Attributes:
        x (vector): State vector (q + dq)
        costs (dict): Costs for this node
        isTerminal (bool): True if last node of shooting problem
        a (vector): Acceleration slack variables (ddq)
        u (vector): Control vector (tau)
        f (vector): Contact forces
        xnext (vector): Next state vector

    Args:
        isTerminal (bool): whether this is a terminal node
    """

    def __init__(self, isTerminal=False):
        self.x = None
        self.costs = {}
        self.isTerminal = isTerminal
        if not isTerminal:
            self.a = None
            self.u = None
            self.f = None
            self.xnext = None


class OcpResult:
    """Result of the OCP

    Attributes:
        xs (vector): State trajectory (q + dq)
        a (vector): Acceleration trajectory (ddq)
        us (vector): Control trajectory (tau)
        fs (vector): Contact forces trajectories
    """

    def __init__(self):
        self.xs = None
        self.a = None
        self.us = None
        self.fs = None


class OCP:
    """Optimal Control Problem

    Args:
        pd (ProblemData): data related to the OCP
        gait (array): array containing the contact pattern i.e. for two steps[[1,1,1,1], [1,0,0,1]]
            It has one column for each feet, the number of rows determines the horizon of the ocp

    Attributes:
        pd (ProblemData): data related to the OCP
        results (OcpResult): result of the OCP
        shooting_nodes (dict): shooting nodes corresponding to possible contact status
        runningModels (list of ShootingNode): running nodes ordered according to gait matrix
        terminalModel (ShootingNode): terminal node of the OCP
        datas (list of NodeData): list of node data, one for each node of the OCP
        opti (casadi.Opti): optimization problem wrapper
        xs (vector): State trajectory (q + dq)
        a (vector): Acceleration slack trajectory
        us (vector): Control trajectory (tau)
        fs (vector): Contact forces trajectories
        cost(casadi.MX): Costs of the OCP
    """

    def __init__(self, pd: ProblemData, gait):

        self.pd = pd
        self.results = OcpResult()

        # Build the types of shooting nodes based on the contact statuses in the gait matrix
        self.shooting_nodes = {
            contacts: ShootingNode(pd, contacts)
            for contacts in set(
                [
                    tuple(self.pd.allContactIds[i] for i in np.nonzero(step == 1)[0])
                    for step in gait[:-1]
                ]
            )
        }

    def solve(self, x0, targets, gait, guess={}):
        """Solve the OCP

        Args:
            x0 (vector): initial state
            targets (list of arrays): targets of the whole-body tasks (lin and ang velocities)
            gait (array): array containing the contact pattern
            guess (dict): dict containing values for the warm start
                In the form: {'xs':array([T+1, n_states]), 'acs':array([T, nv]),
                'us':array([T, nu]), 'fs': [array([T, nv]),]}
        """

        # Assemble the sequence of nodes from the dict of shooting nodes created at initialization
        self.runningModels = [
            self.shooting_nodes[self.pattern_to_id(gait[t])] for t in range(self.pd.T)
        ]
        self.terminalModel = self.shooting_nodes[self.pattern_to_id(gait[self.pd.T])]

        # Create node data for the T running nodes and the terminal node
        self.datas = [NodeData() for _ in range(self.pd.T)] + [NodeData(True)]

        # Create the various optimization variables of the problem and initialize them
        self.make_opti_variables(x0)

        # Assemble the optimization control problem
        self.costs["cost"], eq_constraints = self.make_ocp(targets, gait, x0)

        # Set the costs to be minimized ...
        self.opti.minimize(self.cost)

        # ... subject to these equality constraints
        self.opti.subject_to(eq_constraints == 0)

        # Problem and solver parameters
        p_opts = {"expand": False}
        s_opts = {
            "tol": 1e-4,
            "acceptable_tol": 1e-4,
            "max_iter": 11,
            # "compl_inf_tol": 1e-2,
            # "constr_viol_tol": 1e-2
            # "resto_failure_feasibility_threshold": 1
            # "linear_solver": "ma57"
        }
        self.opti.solver("ipopt", p_opts, s_opts)

        # Warm start the OCP with initial value and guess
        self.warmstart(x0, guess)

        self.log_costs = {}
        for cost in self.datas[0].costs:
            self.log_costs[cost] = []
        self.opti.callback(lambda i: self.my_debug(i))

        # Solve the optimal control problem
        self.opti.solve_limited()

        # Store state of each node after the solving process
        for data in self.datas:
            data.x = self.opti.value(data.x)
            for cost in data.costs:
                data.costs[cost] = self.opti.value(data.costs[cost])
            if not data.isTerminal:
                data.u = self.opti.value(data.u)
                data.xnext = self.opti.value(data.xnext)

        print("-- End of solve() --")
        """from IPython import embed
        embed()"""
    
    def my_debug(self, i):
        print("---" + str(i) + "---")
        for j in range(4):
            print(self.opti.debug.value(self.datas[0].f[j]))

        for cost in self.datas[0].costs:
            self.log_costs[cost].append([self.opti.value(data.costs[cost]) for data in self.datas])
    
        """for data in self.datas:
            for cost in data.costs:
                self.log_costs[cost].append(self.opti.value(data.costs[cost]))"""

        """log_vel_base_iter = []
        log_friction_cone_iter = []
        for data in self.datas:
            log_vel_base_iter.append(self.opti.value(data.cost_vel_track))
            log_friction_cone_iter.append(self.opti.value(data.cost_friction_cone))
        self.log_vel_base.append(log_vel_base_iter)
        self.log_friction_cone.append(log_friction_cone_iter)
        # print("Base vel: ", self.opti.value(self.datas[1].cost_vel_track))
        print("Base vel: ", log_vel_base_iter)"""

        """x_test = []
        for data in self.datas:
            # x_test = x_test.append(self.opti.value(data.x)[19 + 0])
            x_test = x_test.append(data.isTerminal)
        # print(x_test)"""

    def make_opti_variables(self, x0):
        """Create the optimization variables of the OCP and initialize them

        Args:
            x0 (vector): initial state
        """

        # Casadi optimization problem
        opti = casadi.Opti()
        self.opti = opti

        # Optimization variables
        self.acs = [opti.variable(self.pd.nv) for _ in self.runningModels]
        self.us = [opti.variable(self.pd.nu) for _ in self.runningModels]
        self.xs = [opti.variable(self.pd.nx) for _ in (self.runningModels + [self.terminalModel])]
        self.fs = []
        for m in self.runningModels:
            f_tmp = [opti.variable(3) for _ in range(len(m.contactIds))]
            self.fs.append(f_tmp)
        self.fs = self.fs

    def make_ocp(self, targets, gait, x0):
        """Assemble the optimization control problem

        Args:
            targets (list of arrays): targets of the whole-body tasks (lin and ang velocities)
            gait (array): array containing the contact pattern
            x0 (vector): initial state

        Returns:
            totalcost (casadi.MX): Costs of the OCP
            eq_constraints (casadi.MX): Equality constraints of the OCP
        """

        totalcost = 0
        eq = []
        # First running node is in initial state of the robot
        eq.append(self.runningModels[0].difference(self.xs[0], x0))

        # Set targets in xref of ProblemData
        self.pd.xref[self.pd.nq:(self.pd.nq + 3)] = targets[0]
        self.pd.xref[(self.pd.nq + 3):(self.pd.nq + 6)] = targets[1]

        # Gather costs and equality constraints for each running node
        for t in range(self.pd.T):
            # These change every time! Do not use runningModels[0].x to get the state!!!! use xs[0]
            self.runningModels[t].init(
                self.datas[t], self.xs[t], self.acs[t], self.us[t], self.fs[t], False
            )

            xnext, rcost = self.runningModels[t].calc(
                x_ref=self.pd.xref, u_ref=self.pd.uref, target=targets
            )

            # If it is landing
            if (gait[t] != gait[t - 1]).any():
                print("Contact on ", str(self.runningModels[t].contactIds))
                self.runningModels[t].constraint_landing_feet_cost(self.pd.xref)

            # Constraints
            eq.append(self.runningModels[t].constraint_standing_feet_eq())
            eq.append(self.runningModels[t].constraint_dynamics_eq())
            eq.append(
                self.runningModels[t].difference(self.xs[t + 1], xnext)
                - np.zeros(self.pd.ndx)
            )

            totalcost += rcost

        # Commented since we don't want 0 velocities at the end of horizon
        # eq.append(self.xs[self.pd.T][self.pd.nq :])

        # Gather costs and equality constraints for terminal node
        self.terminalModel.init(
            data=self.datas[self.pd.T], x=self.xs[self.pd.T], isTerminal=True
        )
        totalcost += self.terminalModel.calc(x_ref=self.pd.xref)[1]
        totalcost += rcost

        eq_constraints = casadi.vertcat(*eq)

        return totalcost, eq_constraints

    def warmstart(self, x0, guess=None):
        """Warm start the OCP

        Args:
            x0 (vector): initial state
            guess (dict): dict containing values for the warm start
                In the form: {'xs':array([T+1, n_states]), 'acs':array([T, nv]),
                'us':array([T, nu]), 'fs': [array([T, nv]),]}
        """

        for g in guess:
            if guess[g] == []:
                print("No warmstart provided")
                return 0

        try:
            xs_g = np.array(guess["xs"])
            us_g = np.array(guess["us"])
            acs_g = np.array(guess["acs"])
            fs_g = guess["fs"]

            def xdiff(x1, x2):
                return np.concatenate(
                    [
                        pin.difference(
                            self.pd.model, x1[: self.pd.nq], x2[: self.pd.nq]
                        ),
                        x2[self.pd.nq :] - x1[self.pd.nq :],
                    ]
                )

            for x, xg in zip(self.xs, xs_g):
                self.opti.set_initial(x, xg)
            for a, ag in zip(self.acs, acs_g):
                self.opti.set_initial(a, ag)
            for u, ug in zip(self.us, us_g):
                self.opti.set_initial(u, ug)
            for f, fg in zip(self.fs, fs_g):
                fgsplit = np.split(fg, len(f))
                fgc = []
                [fgc.append(f) for f in fgsplit]
                [self.opti.set_initial(f[i], fgc[i]) for i in range(len(f))]
            print("Got warm start")
        except:
            print("Can't load warm start")
            for x in self.xs:
                self.opti.set_initial(x, x0)

    def get_results(self):
        xs_sol = [self.opti.value(x) for x in self.xs]
        us_sol = [self.opti.value(u) for u in self.us]
        acs_sol = [self.opti.value(a) for a in self.acs]
        fsol = {name: [] for name in self.pd.allContactIds}
        fs_world = {name: [] for name in self.pd.allContactIds}
        fsol_to_ws = []
        for t in range(self.pd.T):
            for i, st_foot in enumerate(self.runningModels[t].contactIds):
                fsol[st_foot].append(self.opti.value(self.fs[t][i]))
            for i, sw_foot in enumerate(self.runningModels[t].freeIds):
                fsol[sw_foot].append(np.zeros(3))

            fsol_to_ws.append(
                np.concatenate(
                    [
                        self.opti.value(self.fs[t][j])
                        for j in range(len(self.runningModels[t].contactIds))
                    ]
                )
            )

            pin.framesForwardKinematics(
                self.pd.model, self.pd.rdata, np.array(xs_sol)[t, : self.pd.nq]
            )
            [
                fs_world[foot].append(self.pd.rdata.oMf[foot].rotation @ fsol[foot][t])
                for foot in fs_world
            ]

        for foot in fs_world:
            fs_world[foot] = np.array(fs_world[foot])

        self.results.x = xs_sol
        self.results.a = acs_sol
        self.results.u = us_sol
        self.results.fs = fsol_to_ws
        return self.results.x, self.results.a, self.results.u, self.results.fs, 0

    def get_feet_position(self, xs_sol):
        """Get feet positions over the prediction horizon

        Args:
            xs_sol (array): State trajectory solution of the OCP
                Array of dimension [n_steps, n_states]

        Returns:
            feet_log (dict): dict of position trajectories, one entry per foot
        """

        feet_log = {i: [] for i in self.pd.allContactIds}
        for foot in feet_log:
            tmp = []
            for i in range(len(xs_sol)):
                tmp += [self.terminalModel.feet[foot](xs_sol[i]).full()[:, 0]]
            feet_log[foot] = np.array(tmp)

        return feet_log

    def get_feet_velocity(self, xs_sol):
        """Get feet velocities over the prediction horizon

        Args:
            xs_sol (array): State trajectory solution of the OCP
                Array of dimension [n_steps, n_states]

        Returns:
            feet_log (dict): dict of velocity trajectories, one entry per foot
        """

        feet_log = {i: [] for i in self.pd.allContactIds}
        for foot in feet_log:
            tmp = []
            for i in range(len(xs_sol)):
                tmp += [self.terminalModel.vfeet[foot](xs_sol[i]).full()[:, 0]]
            feet_log[foot] = np.array(tmp)

        return feet_log

    def get_base_log(self, xs_sol):
        """Get base positions over the prediction horizon

        Args:
            xs_sol (array): State trajectory solution of the OCP
                Array of dimension [n_steps, n_states]

        Returns:
            base_pos_log (array): position of the base over the prediction horizon
        """

        base_pos_log = []
        [
            base_pos_log.append(
                self.terminalModel.baseTranslation(xs_sol[i]).full()[:, 0]
            )
            for i in range(len(xs_sol))
        ]
        base_pos_log = np.array(base_pos_log)
        return base_pos_log

    def pattern_to_id(self, gait):
        """Get the key corresponding to a contact status, to be used with the dict of shooting nodes

        Args:
            gait (array): Contact status

        Returns:
            tuple: key corresponding to the contact status
        """

        return tuple(self.pd.allContactIds[i] for i, c in enumerate(gait) if c == 1)
